from typing import Optional
from pydantic import BaseModel, Field


class WrappedTaskObject(BaseModel):
    """Encloses request or response object with additional metadata"""

    message_id: str = Field(
        description="id of originating message; key for istore retrieval"
    )
    data: dict = Field(description="Tool-defined request or response")



from typing import Optional

from nora_lib.interactions.interactions_service import InteractionsService
from nora_lib.interactions.models import ReturnedMessage


class ContextService:
    """
    Save and retrieve task agent context from interaction store
    """

    def __init__(
        self,
        agent_actor_id: str,  # uuid representing this agent in interaction store
        interactions_base_url: str,
        interactions_bearer_token: Optional[str],
        timeout: int = 30,
    ):
        # If no config is provided, load the configuration based on the environment
        self.interactions_service = self._get_interactions_service(
            interactions_base_url, interactions_bearer_token, timeout
        )
        self.agent_actor_id = agent_actor_id

    def _get_interactions_service(self, url, token, timeout) -> InteractionsService:
        return InteractionsService(url, timeout, token)

    def get_message(self, message_id: str) -> str:
        message: ReturnedMessage = self.interactions_service.get_message(message_id)
        if message.annotated_text:
            return message.annotated_text
        else:
            return message.text



"""
Model for interactions to be sent to the interactions service.
"""

import json
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, Tuple
from uuid import UUID

from pydantic import BaseModel, Field, field_serializer, ConfigDict


class Surface(str, Enum):
    SLACK = "Slack"
    WEB = "NoraWebapp"


class Annotation(BaseModel):
    # Need this config to stringify numeric values in attributes.
    # Otherwise, we'll get 'Input should be a valid string' error.
    model_config = ConfigDict(coerce_numbers_to_str=True)

    tag: str
    span: Tuple[int, int]
    attributes: Optional[Dict[str, str]] = None


class AnnotationBatch(BaseModel):
    actor_id: UUID
    message_id: str
    annotations: List[Annotation]

    @field_serializer("actor_id")
    def serialize_actor_id(self, actor_id: UUID):
        return str(actor_id)


class Message(BaseModel):
    message_id: str
    actor_id: UUID
    text: str
    thread_id: str
    channel_id: str
    surface: Surface
    ts: datetime
    annotations: List[Annotation] = Field(default_factory=list)

    @field_serializer("actor_id")
    def serialize_actor_id(self, actor_id: UUID):
        return str(actor_id)

    @field_serializer("ts")
    def serialize_ts(self, ts: datetime):
        return ts.isoformat()

    @staticmethod
    def from_returned_message(message: "ReturnedMessage") -> "Message":
        if message.message_id is None:
            raise ValueError("Message ID is required")
        if message.thread_id is None:
            raise ValueError("Thread ID is required")
        if message.channel_id is None:
            raise ValueError("Channel ID is required")
        if message.surface is None:
            raise ValueError("Surface is required")
        return Message(
            message_id=message.message_id,
            actor_id=message.actor_id,
            text=message.text,
            thread_id=message.thread_id,
            channel_id=message.channel_id,
            surface=message.surface,
            ts=message.ts,
            annotations=message.annotations,
        )


class Event(BaseModel):
    """event object to be sent to the interactions service; requires association with a message, thread or channel id"""

    type: str
    actor_id: UUID = Field(
        description="identifies actor writing the event to the interaction service"
    )
    timestamp: datetime
    text: Optional[str] = None
    data: dict = Field(default_factory=dict)
    message_id: Optional[str] = None
    thread_id: Optional[str] = None
    channel_id: Optional[str] = None

    @field_serializer("actor_id")
    def serialize_actor_id(self, actor_id: UUID):
        return str(actor_id)

    @field_serializer("timestamp")
    def serialize_timestamp(self, timestamp: datetime):
        return timestamp.isoformat()

    @staticmethod
    def from_returned_event(event: "ReturnedEvent") -> "Event":
        if event.channel_id is None:
            raise ValueError("Channel ID is required")
        elif event.thread_id is None and event.message_id is not None:
            raise ValueError("Thread ID is required if Message ID is present")
        return Event(
            type=event.type,
            actor_id=event.actor_id,
            timestamp=event.timestamp,
            text=event.text,
            data=event.data,
            message_id=event.message_id,
            thread_id=event.thread_id,
            channel_id=event.channel_id,
        )


class EventType(Enum):
    """Enumeration of event types"""

    STEP_COST = "step_cost"
    STEP_PROGRESS = "step_progress"


class Thread(BaseModel):
    thread_id: str
    channel_id: str
    surface: Surface


class ReturnedEvent(BaseModel):
    """Event format returned by the interaction service"""

    event_id: str
    type: str
    actor_id: UUID = Field(
        description="identifies actor writing the event to the interaction service"
    )
    timestamp: datetime
    text: Optional[str] = None
    data: dict = Field(default_factory=dict)
    message_id: Optional[str] = None
    thread_id: Optional[str] = None
    channel_id: Optional[str] = None
    surface: Optional[Surface] = None

    @field_serializer("actor_id")
    def serialize_actor_id(self, actor_id: UUID):
        return str(actor_id)

    @field_serializer("timestamp")
    def serialize_timestamp(self, timestamp: datetime):
        return timestamp.isoformat()


class ReturnedMessage(BaseModel):
    """Message format returned by interaction service"""

    actor_id: UUID
    text: str
    ts: datetime
    message_id: Optional[str] = None
    annotated_text: Optional[str] = None
    events: List[Event] = Field(default_factory=list)
    preceding_messages: List["ReturnedMessage"] = Field(default_factory=list)
    thread_id: Optional[str] = None
    channel_id: Optional[str] = None
    surface: Optional[Surface] = None
    annotations: List[Annotation] = Field(default_factory=list)

    @classmethod
    def from_event(cls, event: Event) -> "ReturnedMessage":
        """Convert an event to a message"""
        return ReturnedMessage(
            actor_id=event.actor_id,
            text=json.dumps(event.data),
            ts=event.timestamp,
            message_id=event.message_id,
        )


class AgentMessageData(BaseModel):
    """capture requests to and responses from tools within Events"""

    message_data: dict  # dict of agent/tool request/response format
    data_sender_actor_id: Optional[str] = None  # agent sending the data
    virtual_thread_id: Optional[str] = None  # tool-provided thread
    tool_call_id: Optional[str] = None  # llm-provided thread
    tool_name: Optional[str] = None  # llm identifier for tool


class ReturnedAgentContextEvent(BaseModel):
    """Event format returned by interaction service for agent context events"""

    actor_id: UUID  # agent that saved this context
    timestamp: datetime
    data: AgentMessageData
    type: str


class ReturnedAgentContextMessage(BaseModel):
    """Message format returned by interaction service for search by thread"""

    message_id: str
    actor_id: UUID
    text: str
    ts: str
    annotated_text: Optional[str] = None
    events: List[ReturnedAgentContextEvent] = Field(default_factory=list)


class ThreadRelationsResponse(BaseModel):
    """Thread format returned by interaction service for thread relations in a search response"""

    thread_id: str
    events: List[Event] = Field(
        default_factory=list
    )  # events associated only with the thread
    messages: List[ReturnedMessage] = Field(
        default_factory=list
    )  # includes events associated with each message


class VirtualThread:
    """Virtuals threads are an event type used to sub-divide a thread into sb-conversations"""

    # The type of event that represetns a virtual thread
    EVENT_TYPE = "virtual_thread"

    # Data field in the event that contains the ID of the virtual thread id
    ID_FIELD = "virtual_thread_id"

    # Data field in the event that contains the type of other events in the virtual thread
    EVENT_TYPE_FIELD = "event_type"


class CostDetail(BaseModel):
    """
    Base class to store details of cost to service a request by an agent.
    If an agent has different cost details,
    it should create another class inheriting this class and add those fields.
    See LLMCost and LLMTokenBreakdown below for examples.
    """

    model_config = ConfigDict(protected_namespaces=())
    pass


class LLMCost(CostDetail):
    """LLM cost detail"""

    token_count: int
    model_name: str


class LLMTokenBreakdown(CostDetail):
    """Token usage breakdown"""

    prompt_tokens: int
    completion_tokens: int


class LangChainRun(CostDetail):
    """LangChain Run"""

    # Subset of run fields which allow future lookup of run details.
    run_id: UUID
    run_name: Optional[str] = None
    trace_id: Optional[UUID] = None
    session_name: Optional[str] = None  # Alias: project_name
    session_id: Optional[UUID] = None  # Alias: project_id

    # Serialize the UUIDs as strings
    @field_serializer("run_id")
    def serialize_id(self, run_id: UUID):
        return str(run_id)

    @field_serializer("trace_id")
    def serialize_trace_id(self, trace_id: UUID):
        return str(trace_id)

    @field_serializer("session_id")
    def serialize_session_id(self, session_id: UUID):
        return str(session_id)


class ServiceCost(BaseModel):
    """Cost of servicing a request by an agent"""

    dollar_cost: float
    service_provider: Optional[str] = Field(
        default=None, description="For example, OpenAI/Anthropic/Modal/Cohere"
    )
    description: Optional[str] = Field(
        default=None, description="Describe the function within the agent"
    )
    tool_name: Optional[str] = Field(default=None, description="For example, PaperQA")
    task_id: Optional[str] = Field(
        default=None,
        description="Agent generated task_id used to track nora assigned tasks",
    )
    tool_call_id: Optional[str] = None
    details: List[CostDetail] = []


class StepCost(BaseModel):
    """Wrapping service cost with event metadata so that it can be converted to an Event object."""

    actor_id: UUID
    message_id: str
    service_cost: ServiceCost

    @field_serializer("actor_id")
    def serialize_actor_id(self, actor_id: UUID):
        return str(actor_id)

    def to_event(self) -> Event:
        return Event(
            type=EventType.STEP_COST.value,
            actor_id=self.actor_id,
            timestamp=datetime.now(),
            text=self.service_cost.description,
            # This flag is needed to serialize subclass
            # https://docs.pydantic.dev/latest/concepts/serialization/#serializeasany-annotation
            data=self.service_cost.model_dump(serialize_as_any=True),
            message_id=self.message_id,
        )


def thread_message_lookup_request(message_id: str, event_type: str) -> dict:
    """retrieve messages and events for the thread associated with a message"""
    return {
        "id": message_id,
        "relations": {
            "thread": {
                "relations": {
                    "messages": {
                        "relations": {"events": {"filter": {"type": event_type}}},
                        "apply_annotations_from_actors": ["*"],
                    },
                }
            }
        },
    }



from datetime import datetime
from uuid import UUID

from pydantic import PlainSerializer
from typing_extensions import Annotated

UuidWithSerializer = Annotated[UUID, PlainSerializer(lambda x: str(x), return_type=str)]
DatetimeWithSerializer = Annotated[
    datetime, PlainSerializer(lambda x: x.isoformat(), return_type=str)
]



from typing import Dict, Any

from nora_lib.interactions.models import Surface
from pydantic import BaseModel, Field


class MessageAgentContext(BaseModel):
    """
    Identifiers for the triggering user message
    """

    message_id: str
    thread_id: str
    channel_id: str
    surface: Surface


class PubsubAgentContext(BaseModel):
    """
    The pubsub namespace in which the Handler is running
    """

    base_url: str
    namespace: str


class ToolConfigAgentContext(BaseModel):
    """
    The name of the tool config being used by the handler (handler-best-quality, low-cost, etc.)
    """

    config_name: str
    overrides: Dict[str, Any] = Field(default_factory=dict)


class AgentContext(BaseModel):
    """
    Information that needs to be passed from the Handler to tool agents
    """

    message: MessageAgentContext
    pubsub: PubsubAgentContext
    tool_config: ToolConfigAgentContext



from typing import Any, Dict, Generic, Optional, TypeVar
from pydantic import BaseModel, Field


R = TypeVar("R", bound=BaseModel)


TASK_STATUSES = dict(STARTED="STARTED", FAILED="FAILED", COMPLETED="COMPLETED")


class AsyncTaskState(BaseModel, Generic[R]):
    """Models the current state of an asynchronous request."""

    task_id: str = Field(
        "Identifies the long-running task so that its status and eventual result"
        "can be checked in follow-up calls."
    )
    estimated_time: str = Field(
        description="How long we expect this task to take from start to finish."
    )
    task_status: str = Field(description="Current human-readable status of the task.")
    task_result: Optional[R] = Field(description="Final result of the task.")
    extra_state: Dict[str, Any] = Field(
        description="Any extra task-specific state can go in here as free-form JSON-serializable dictionary."
    )



import logging

import requests
from typing import Optional, List
import json
import os
import boto3
from requests import Response
from requests.auth import AuthBase
from aws_requests_auth.aws_auth import AWSRequestsAuth
from typing import Dict, Any

from nora_lib.interactions.models import (
    AnnotationBatch,
    StepCost,
    Event,
    Message,
    ReturnedMessage,
    ReturnedEvent,
    Thread,
    ThreadRelationsResponse,
    VirtualThread,
)


class InteractionsService:
    """
    Service which saves interactions to the Interactions API
    """

    def __init__(
        self,
        base_url: str,
        timeout: int = 30,
        token: Optional[str] = None,
        auth: Optional[AuthBase] = None,
    ) -> None:
        self.base_url = base_url
        self.timeout = timeout
        if auth:
            self.auth = auth
        elif token:
            self.auth = BearerAuth(token)
        else:
            raise Exception("Either token or auth must be provided")

    def _post(self, url: str, json: Dict[str, Any]) -> Response:
        return requests.post(
            url,
            json=json,
            auth=self.auth,
            timeout=self.timeout,
        )

    def save_message(
        self, message: Message, virtual_thread_id: Optional[str] = None
    ) -> None:
        """
        Save a message to the Interaction Store
        :param virtual_thread_id: Optional ID of a virtual thread to associate with the message
        """
        message_url = f"{self.base_url}/interaction/v1/message"
        response = self._post(
            message_url,
            message.model_dump(),
        )
        response.raise_for_status()
        if virtual_thread_id:
            # Use an event to tag the message with the virtual thread ID
            event = Event(
                type=VirtualThread.EVENT_TYPE,
                actor_id=message.actor_id,
                message_id=message.message_id,
                data={
                    VirtualThread.ID_FIELD: virtual_thread_id,
                    VirtualThread.EVENT_TYPE_FIELD: VirtualThread.EVENT_TYPE,
                },
                timestamp=message.ts,
            )
            self.save_event(event)

    def save_event(self, event: Event, virtual_thread_id: Optional[str] = None) -> str:
        """
        Save an event to the Interaction Store. Returns an event id.
        :param virtual_thread_id: Optional ID of a virtual thread to associate with the event
        """
        event_url = f"{self.base_url}/interaction/v1/event"
        response = self._post(
            event_url,
            event.model_dump(),
        )
        response.raise_for_status()
        if virtual_thread_id:
            # Use an event to tag the event with the virtual thread ID
            # Attach it to the same message as this event, along with the event type
            event = Event(
                type=VirtualThread.EVENT_TYPE,
                actor_id=event.actor_id,
                message_id=event.message_id,
                data={
                    VirtualThread.ID_FIELD: virtual_thread_id,
                    VirtualThread.EVENT_TYPE_FIELD: event.type,
                },
                timestamp=event.timestamp,
            )
            self.save_event(event)
        response_message = json.loads(response.text)
        event_id = response_message["event_id"]
        return event_id

    def save_thread(self, thread: Thread) -> None:
        """Save a thread to the Interactions API"""
        thread_url = f"{self.base_url}/interaction/v1/thread"
        response = self._post(
            thread_url,
            thread.model_dump(),
        )
        response.raise_for_status()

    def get_virtual_thread_content(
        self, message_id: str, virtual_thread_id: str
    ) -> List[ReturnedMessage]:
        """Fetch all messages and events in a virtual thread
        Returns all messages and events in the same thread as the given message,
        but filtered to only include those associated with the given virtual thread.
        :param message_id: The ID of a message in the virtual thread
        :param virtual_thread_id: The ID of the virtual thread
        """
        message_search_url = f"{self.base_url}/interaction/v1/search/message"
        # Fetch all events and filter on the client side
        # Need an IStore schema change to do this server-side
        request_body = {
            "id": message_id,
            "relations": {
                "preceding_messages": {
                    "max": 100,
                    "relations": {"events": {}},
                },
                "events": {},
            },
        }

        response = self._post(
            message_search_url,
            request_body,
        )
        response.raise_for_status()
        result = ReturnedMessage.model_validate(response.json()["message"])
        all_messages = result.preceding_messages + [result]
        virtual_thread_content = []
        for msg in all_messages:
            event_types_in_virtual_thread = set(
                event.data[VirtualThread.EVENT_TYPE_FIELD]
                for event in msg.events
                if event.type == VirtualThread.EVENT_TYPE
                and event.data.get(VirtualThread.ID_FIELD) == virtual_thread_id
            )
            if not event_types_in_virtual_thread:
                continue
            virtual_thread_content.append(msg)
            msg.events = [
                event
                for event in msg.events
                if event.type != VirtualThread.EVENT_TYPE
                and event.type in event_types_in_virtual_thread
            ]
            if VirtualThread.EVENT_TYPE not in event_types_in_virtual_thread:
                # An event has been tagged with the virtual thread ID
                # but the message itself is not in the virtual thread
                # Somewhat pathological case, probably shouldn't happen
                # Set the message text to empty string
                msg.text = ""
        return virtual_thread_content

    def save_annotation(self, annotation: AnnotationBatch) -> None:
        """Save an annotation to the Interactions API"""
        annotation_url = f"{self.base_url}/interaction/v1/annotation"
        response = self._post(
            annotation_url,
            annotation.model_dump(),
        )
        response.raise_for_status()

    def get_message(self, message_id: str) -> ReturnedMessage:
        """Fetch a message from the Interactions API"""
        message_url = f"{self.base_url}/interaction/v1/search/message"
        request_body = {
            "id": message_id,
            "relations": {"thread": {}, "channel": {}, "events": {}, "annotations": {}},
        }
        response = self._post(
            message_url,
            request_body,
        )
        response.raise_for_status()
        res_dict = response.json()["message"]
        res = ReturnedMessage.model_validate(res_dict)

        # thread_id and channel_id are for some reason nested in the response
        if not res.thread_id:
            res.thread_id = res_dict.get("thread", {}).get("thread_id")
        if not res.channel_id:
            res.channel_id = res_dict.get("channel", {}).get("channel_id")

        return res

    def get_event(self, event_id: str) -> ReturnedEvent:
        """Fetch an event from the Interactions API"""
        event_url = f"{self.base_url}/interaction/v1/search/event"
        request_body = {
            "id": event_id,
        }
        response = self._post(
            event_url,
            request_body,
        )
        response.raise_for_status()
        res_dict = response.json()["events"][0]
        res = ReturnedEvent.model_validate(res_dict)
        return res

    def fetch_all_threads_by_channel(
        self,
        channel_id: str,
        min_timestamp: Optional[str] = None,
        thread_event_types: Optional[list[str]] = None,
        most_recent: Optional[int] = None,
    ) -> dict:
        """Fetch a message from the Interactions API"""
        message_url = f"{self.base_url}/interaction/v1/search/channel"
        request_body = self._channel_lookup_request(
            channel_id=channel_id,
            min_timestamp=min_timestamp,
            thread_event_types=thread_event_types,
            most_recent=most_recent,
        )
        response = self._post(
            message_url,
            request_body,
        )
        response.raise_for_status()
        return response.json()

    def fetch_thread_messages_and_events_for_message(
        self,
        message_id: str,
        event_types: List[str],
        min_timestamp: Optional[str] = None,
        most_recent: Optional[int] = None,
    ) -> ThreadRelationsResponse:
        """Fetch messages sorted by timestamp and events for agent context"""
        message_url = f"{self.base_url}/interaction/v1/search/message"
        request_body = self._thread_lookup_request(
            message_id,
            event_types=event_types,
            min_timestamp=min_timestamp,
            most_recent=most_recent,
        )
        response = self._post(
            message_url,
            request_body,
        )
        response.raise_for_status()
        json_response = response.json()

        return ThreadRelationsResponse.model_validate(
            json_response.get("message", {}).get("thread", {})
        )

    def fetch_messages_and_events_for_thread(
        self,
        thread_id: str,
        event_type: Optional[str] = None,
        min_timestamp: Optional[str] = None,
    ) -> dict:
        """Fetch messages and events for the given thread from the Interactions API"""
        thread_search_url = f"{self.base_url}/interaction/v1/search/thread"
        message_query = {
            "filter": {"min_timestamp": min_timestamp} if min_timestamp else None,
            "apply_annotations_from_actors": ["*"],
        }
        request_body = {
            "id": thread_id,
            "relations": {
                "messages": message_query,
                "events": {"filter": {"type": event_type}} if event_type else {},
            },
        }

        response = self._post(
            thread_search_url,
            request_body,
        )
        response.raise_for_status()
        return response.json()

    def fetch_events_for_message(
        self,
        message_id: str,
        event_type: Optional[str] = None,
    ) -> dict:
        """Fetch messages and events for the thread containing a given message from the Interactions API"""
        message_search_url = f"{self.base_url}/interaction/v1/search/message"
        request_body = {
            "id": message_id,
            "relations": {
                "events": {"filter": {"type": event_type}} if event_type else {},
            },
        }

        response = self._post(
            message_search_url,
            request_body,
        )
        response.raise_for_status()
        return response.json()

    def fetch_all_by_channel(
        self,
        channel_id: str,
        min_timestamp: Optional[str] = None,
        event_types: Optional[List[str]] = None,
        most_recent: Optional[int] = None,
    ) -> dict:
        """
        Fetch all threads, messages, and events including nested ones for a given channel
        """
        channel_search_url = f"{self.base_url}/interaction/v1/search/channel"
        event_query = {"filter": None if event_types is None else {"type": event_types}}
        message_filter_query = {
            "min_timestamp": min_timestamp if min_timestamp else None,
            "most_recent": most_recent if most_recent else None,
        }
        message_query = {
            "relations": {"events": event_query, "annotations:": {}},
            "filter": message_filter_query,
            "apply_annotations_from_actors": ["*"],
        }
        request_body = {
            "id": channel_id,
            "relations": {
                "threads": {
                    "relations": {
                        "messages": message_query,
                        "events": event_query,
                    }
                },
                "messages": message_query,
            },
        }
        response = self._post(
            channel_search_url,
            request_body,
        )
        response.raise_for_status()
        return response.json()

    def fetch_all_by_thread(
        self,
        thread_id: str,
        min_timestamp: Optional[str] = None,
        event_types: Optional[List[str]] = None,
        most_recent: Optional[int] = None,
    ) -> dict:
        """
        Fetch all messages and events including nested ones for a given thread
        """
        thread_search_url = f"{self.base_url}/interaction/v1/search/thread"
        event_query = {"filter": None if event_types is None else {"type": event_types}}
        message_filter_query = {
            "min_timestamp": min_timestamp if min_timestamp else None,
            "most_recent": most_recent if most_recent else None,
        }
        message_query = {
            "relations": {"events": event_query, "annotations:": {}},
            "filter": message_filter_query,
            "apply_annotations_from_actors": ["*"],
        }
        request_body = {
            "id": thread_id,
            "relations": {
                "messages": message_query,
                "events": event_query,
            },
        }
        response = self._post(
            thread_search_url,
            request_body,
        )
        response.raise_for_status()
        return response.json()

    def report_cost(self, step_cost: StepCost) -> Optional[str]:
        """Save a cost report to the Interactions Store. Returning event id"""
        try:
            return self.save_event(step_cost.to_event())
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logging.warning(
                    f"Cannot find message id {step_cost.message_id} to attach cost report to."
                )
                return None
            else:
                raise e

    @staticmethod
    def _channel_lookup_request(
        channel_id: str,
        min_timestamp: Optional[str] = None,
        thread_event_types: Optional[list[str]] = None,
        most_recent: Optional[int] = None,
    ) -> dict:
        """Interaction service API request to get threads and messages for a channel"""
        message_filter_query = {
            "min_timestamp": min_timestamp if min_timestamp else None,
            "most_recent": most_recent if most_recent else None,
        }
        return {
            "id": channel_id,
            "relations": {
                "threads": {
                    "relations": {
                        "messages": {
                            "filter": message_filter_query,
                            "apply_annotations_from_actors": ["*"],
                        },
                        "events": {"filter": {"type": thread_event_types or []}},
                    }
                }
            },
        }

    @staticmethod
    def _thread_lookup_request(
        message_id: str,
        event_types: list[str],
        min_timestamp: Optional[str] = None,
        most_recent: Optional[int] = None,
    ) -> dict:
        """will return all messages for the thread containing the given message and events associated with each message"""
        message_filter_query = {
            "min_timestamp": min_timestamp if min_timestamp else None,
            "most_recent": most_recent if most_recent else None,
        }
        return {
            "id": message_id,
            "relations": {
                "thread": {
                    "relations": {
                        "messages": {
                            "filter": message_filter_query,
                            "relations": {"events": {"filter": {"type": event_types}}},
                            "apply_annotations_from_actors": ["*"],
                        },
                    }
                }
            },
        }

    @staticmethod
    def fetch_bearer_token(secret_id: str) -> str:
        secrets_manager = boto3.client("secretsmanager", region_name="us-west-2")
        return json.loads(
            secrets_manager.get_secret_value(SecretId=secret_id)["SecretString"]
        )["token"]

    @staticmethod
    def from_env() -> "InteractionsService":
        """Load the configuration based on the environment."""
        use_api_gw = (
            os.getenv("INTERACTION_STORE_VIA_API_GW", "false").lower() == "true"
        )
        if use_api_gw:
            return InteractionsService.via_api_gw()
        else:
            url = os.getenv(
                "INTERACTION_STORE_URL",
                "https://s2ub.prod.s2.allenai.org/service/noraretrieval",
            )
            token = os.getenv(
                "INTERACTION_STORE_TOKEN",
                InteractionsService.fetch_bearer_token(
                    "nora/prod/interaction-bearer-token"
                ),
            )

        return InteractionsService(base_url=url, token=token)

    @staticmethod
    def via_api_gw() -> "InteractionsService":
        credentials = boto3.Session().get_credentials()
        region = "us-west-2"
        service = "execute-api"
        host = f"95e441vq07.{service}.{region}.amazonaws.com"
        url = f"https://{host}/prod/nora-interactions"

        auth = AWSRequestsAuth(
            aws_access_key=credentials.access_key,
            aws_secret_access_key=credentials.secret_key,
            aws_token=credentials.token,
            aws_host=host,
            aws_region="us-west-2",
            aws_service="execute-api",
        )
        return InteractionsService(base_url=url, auth=auth)


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r



"""
Handles IO for asynchronous task-related state.

A StateManager instance must be initialized with
a concrete subclass of `AsyncTaskState`, as implemented
by dependent projects.
"""

import json
import logging
import os
from uuid import UUID
from datetime import datetime, timezone
from typing import Generic, Optional, Type, Any
from abc import ABC, abstractmethod
from pydantic import BaseModel

from nora_lib.tasks.models import AsyncTaskState, R, TASK_STATUSES
from nora_lib.interactions.interactions_service import InteractionsService
from nora_lib.interactions.models import Event, ReturnedEvent
from nora_lib.pubsub import PubsubService
from nora_lib.context.agent_context import AgentContext

TASK_STATE_CHANGE_TOPIC = "istore:event:task_state"


class NoSuchTaskException(Exception):
    def __init__(self, task_id: str):
        self._task_id = task_id

    def __str__(self):
        return f"No record found for task {self._task_id}"


class IStateManager(ABC, Generic[R]):
    @abstractmethod
    def read_state(self, task_id: str) -> AsyncTaskState[R]:
        pass

    @abstractmethod
    def write_state(self, state: AsyncTaskState[R]) -> None:
        pass

    def update_status(self, task_id: str, new_status: str) -> None:
        state = self.read_state(task_id)
        state.task_status = new_status
        self.write_state(state)

    def save_result(self, task_id: str, task_result: R) -> None:
        state = self.read_state(task_id)
        state.task_status = TASK_STATUSES["COMPLETED"]
        state.task_result = task_result
        self.write_state(state)


class StateManager(IStateManager[R]):
    """
    Stores task state on local disk
    """

    def __init__(self, task_state_class: Type[AsyncTaskState[R]], state_dir) -> None:
        self._task_state_class = task_state_class
        self._state_dir = state_dir

    def read_state(self, task_id: str) -> AsyncTaskState[R]:
        task_state_path = os.path.join(self._state_dir, f"{task_id}.json")
        if not os.path.isfile(task_state_path):
            raise NoSuchTaskException(task_id)

        with open(task_state_path, "r") as f:
            return self._task_state_class(**json.loads(f.read()))

    def write_state(self, state: AsyncTaskState[R]) -> None:
        task_state_path = os.path.join(self._state_dir, f"{state.task_id}.json")
        with open(task_state_path, "w") as f:
            json.dump(state.model_dump(), f)


class TaskStateFetchException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class RemoteStateManagerFactory:
    """
    Stores task state in the interaction store
    """

    def __init__(
        self,
        agent_name: str,
        actor_id: UUID,
        interactions_service: InteractionsService,
        pubsub_service: PubsubService,
    ):
        """
        :param agent_name: Used to form the event type that will hold the task state in the interactions store
        :param actor_id: Associated with the events written to the interactions store
        :param interactions_service:
        """
        self.agent_name = agent_name
        self.actor_id = actor_id
        self.interactions_service = interactions_service
        self.pubsub_service = pubsub_service

    def for_message(self, message_id: str) -> IStateManager[R]:
        return RemoteStateManager(
            self.agent_name,
            self.actor_id,
            self.interactions_service,
            self.pubsub_service,
            message_id,
        )

    def for_agent_context(self, context: AgentContext) -> IStateManager[R]:
        return RemoteStateManager(
            self.agent_name,
            self.actor_id,
            self.interactions_service,
            PubsubService(context.pubsub.base_url, context.pubsub.namespace),
            context.message.message_id,
        )


class RemoteStateManager(IStateManager[R]):
    """
    Stores task state in the interaction store
    """

    _TASK_STATE_EVENT_TYPE = "agent:{}:task_state"

    def __init__(
        self,
        agent_name: str,
        actor_id: UUID,
        interactions_service: InteractionsService,
        pubsub_service: PubsubService,
        message_id: str,
    ):
        """
        :param agent_name: Agent that saved the task
        :param actor_id: ID for the agent (ignored when reading)
        :param message_id: The message that initiated the request for task status
        """
        self.agent_name = agent_name
        self.actor_id = actor_id
        self.message_id = message_id
        self.interactions_service = interactions_service
        self.pubsub_service = pubsub_service

    def read_state(self, task_id: str) -> AsyncTaskState[R]:
        event_type = RemoteStateManager._TASK_STATE_EVENT_TYPE.format(self.agent_name)
        response = (
            self.interactions_service.fetch_thread_messages_and_events_for_message(
                self.message_id, [event_type]
            )
        )
        latest_state: Optional[AsyncTaskState[R]] = None
        latest_timestamp = None
        for msg in response.messages or []:
            for event in msg.events or []:
                try:
                    state = AsyncTaskState[Any].model_validate(event.data)
                except Exception as e:
                    # Event json blob has unexpected format
                    raise TaskStateFetchException(
                        f"Event of type {event_type} for message {self.message_id} does not deserialize to AsyncTaskState: {e}"
                    )
                if state.task_id != task_id:
                    continue
                if latest_state is None or (
                    latest_timestamp and event.timestamp > latest_timestamp
                ):
                    latest_state = state
                    latest_timestamp = event.timestamp

        if not latest_state:
            raise NoSuchTaskException(task_id)
        return latest_state

    def write_state(self, state: AsyncTaskState[R]) -> None:
        event_type = RemoteStateManager._TASK_STATE_EVENT_TYPE.format(self.agent_name)
        event = Event(
            type=event_type,
            actor_id=self.actor_id,
            timestamp=datetime.now(tz=timezone.utc),
            message_id=self.message_id,
            data=state.model_dump(),
        )
        event_id = self.interactions_service.save_event(event)
        returned_event = ReturnedEvent(
            event_id=event_id,
            type=event.type,
            actor_id=event.actor_id,
            message_id=event.message_id,
            timestamp=event.timestamp,
        )
        payload = TaskStateChangeNotification(
            agent=self.agent_name, event=returned_event
        )
        try:
            self.pubsub_service.publish(TASK_STATE_CHANGE_TOPIC, payload.model_dump())
        except Exception as e:
            logging.exception(
                f"Failed to publish event to pubsub topic {TASK_STATE_CHANGE_TOPIC} at {self.pubsub_service.base_url}"
            )


class TaskStateChangeNotification(BaseModel):
    agent: str
    event: ReturnedEvent



import logging
import uuid
from datetime import datetime, timezone
from enum import Enum
from typing import Optional
from uuid import UUID

import requests
from pydantic import BaseModel

from nora_lib.interactions.serializers import UuidWithSerializer, DatetimeWithSerializer
from nora_lib.interactions.interactions_service import InteractionsService
from nora_lib.interactions.models import Event, EventType


class RunState(str, Enum):
    """State of a step"""

    CREATED = "created"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class StepProgress(BaseModel):
    """Data class for step progress. This goes into the `data` field in `Event`."""

    # A short message, e.g. "searching for $query". Recommend < 100 chars.
    short_desc: str
    # Detailed message.
    long_desc: Optional[str] = None
    # Updates on the same unit of work have the same step_id.
    step_id: UuidWithSerializer = uuid.uuid4()
    # Inner steps can be constituent to some outer step, effectively a tree.
    parent_step_id: Optional[UuidWithSerializer] = None
    # Populated if this step is due to an async task.
    task_id: Optional[str] = None

    # Enum of possible states.
    run_state: RunState = RunState.CREATED
    # DB timestamp when this step was defined/created.
    created_at: Optional[DatetimeWithSerializer] = None
    # When this step started running.
    started_at: Optional[DatetimeWithSerializer] = None
    # Estimated finish time, if available.
    finish_est: Optional[DatetimeWithSerializer] = None
    # When this step stopped running, whether that was due to success or failure.
    finished_at: Optional[DatetimeWithSerializer] = None
    # Error message in case of terminal step failure.
    error_message: Optional[str] = None


class StepProgressReporter:
    """
    Wrapper around StepProgress to add event metadata and report to interactions service

    Usage:
    # Create/define a step
    find_papers_progress = StepProgressReporter(
        actor_id,
        message_id,
        StepProgress(short_desc="Find papers"),
        interactions_service
    )

    # Do something
    ...

    # Start the step
    find_papers_progress.start()

    # Do more things
    ...
    count_citation_progress = find_papers_progress.create_child_step(short_desc="Count citations")
    ...
    count_citation_progress.start()
    ...
    count_citation_progress.finish(is_success=True)
    ...

    # Finish the step
    find_papers_progress.finish(is_success=False, error_message="Something went wrong")

    # Alternatively, you can use this as a context.

    with StepProgressReporter(...) as spr:
        # Do something
        ...

        # You still need to start the step
        spr.start()

    # This step will be automatically finished when the context exits.
    # If an exception is raised, the step will be marked as failed
    # and the exception message will be recorded in the error_message field
    """

    def __init__(
        self,
        actor_id: UUID,
        message_id: str,
        step_progress: StepProgress,
        interactions_service: InteractionsService,
    ):
        self.actor_id = actor_id
        self.message_id = message_id
        self.step_progress = step_progress
        self.interactions_service = interactions_service

        # Report step creation
        self.step_progress.run_state = RunState.CREATED
        event_id_opt = self._report_progress()

        # Use DB timestamp for created_at
        if event_id_opt:
            event = self.interactions_service.get_event(event_id_opt)
            self.step_progress.created_at = event.timestamp

    def __enter__(self):
        return self

    def __exit__(self, error_type, value, traceback):
        is_success = error_type is None
        self.finish(is_success=is_success, error_message=str(value))
        return True

    def start(self):
        """Start a step"""
        if self.step_progress.run_state in [
            RunState.RUNNING,
            RunState.SUCCEEDED,
            RunState.FAILED,
        ]:
            logging.warning(
                f"Trying to start an already started/finished step. "
                f"Doing nothing instead. Step id: {self.step_progress.step_id}. "
                f"Run state: {self.step_progress.run_state}."
            )
        else:
            self.step_progress.started_at = datetime.now(timezone.utc)
            self.step_progress.run_state = RunState.RUNNING
            self._report_progress()

    def finish(self, is_success: bool, error_message: Optional[str] = None):
        """Finish a step whether it was successful or not"""
        if self.step_progress.run_state in [RunState.SUCCEEDED, RunState.FAILED]:
            logging.warning(
                f"Trying to finish an already finished step. "
                f"Doing nothing instead. Step id: {self.step_progress.step_id}. "
                f"Run state: {self.step_progress.run_state}."
            )
        elif self.step_progress.run_state != RunState.RUNNING:
            logging.warning(
                f"Trying to finish a step that has not been started. "
                f"Doing nothing instead. Step id: {self.step_progress.step_id}"
            )
        else:
            self.step_progress.finished_at = datetime.now(timezone.utc)
            self.step_progress.run_state = (
                RunState.SUCCEEDED if is_success else RunState.FAILED
            )
            self.step_progress.error_message = error_message if error_message else None
            self._report_progress()

    def create_child_step(
        self, short_desc: str, long_desc: Optional[str] = None
    ) -> "StepProgressReporter":
        """Create a child step"""
        child_step_progress_event = StepProgressReporter(
            actor_id=self.actor_id,
            message_id=self.message_id,
            step_progress=StepProgress(
                parent_step_id=self.step_progress.step_id,
                short_desc=short_desc,
                long_desc=long_desc,
            ),
            interactions_service=self.interactions_service,
        )
        return child_step_progress_event

    def _report_progress(self) -> Optional[str]:
        """Report a step progress to the Interactions Store. Returns the event id if successful."""
        try:
            return self.interactions_service.save_event(self._to_event())
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logging.warning(
                    f"Cannot find message id {self.message_id} to attach step progress to."
                )
                return None
            else:
                raise e

    def _to_event(self) -> Event:
        return Event(
            type=EventType.STEP_PROGRESS.value,
            actor_id=self.actor_id,
            timestamp=datetime.now(),
            data=self.step_progress.model_dump(),
            message_id=self.message_id,
        )



import logging
from typing import Dict, Any, Iterator, Optional
from contextlib import contextmanager

import os
import json
import requests
import signal
from time import sleep

from pydantic import BaseModel


class PubsubService:
    """Client for Nora pubsub backend"""

    def __init__(self, base_url: str, namespace: Optional[str] = None):
        """
        :param base_url: pubsub API URL
        :param namespace: Topic namespace
        """
        self.base_url = base_url
        self.namespace = namespace

    def subscribe_webhook(self, topic: str, url: str):
        """
        Add a webhook subscriber to a topic
        The webhook will receive a POST request whenever any client calls publish() on the topic
        """
        body = {"url": url}
        requests.post(
            f"{self.base_url}/subscribe/webhook/{self._fully_qualified_topic(topic)}",
            json=body,
        )

    def unsubscribe_webhook(self, topic: str, url: str):
        """
        Remove a webhook subscriber from a topic
        """
        body = {"url": url}
        requests.post(
            f"{self.base_url}/unsubscribe/webhook/{self._fully_qualified_topic(topic)}",
            json=body,
        )

    @contextmanager
    def subscribe_sse(self, topic: str) -> Iterator[str]:
        """
        Subscribe to a topic using Server-Sent Events
        Returns an iterator that yields message payloads as they are published
        Will close the underlying HTTP connection when the context manager exits

        Usage:

        # This will run indefinitely
        with pubsub_service.subscribe_sse("my_topic") as messages:
             for message in messages:
                    handle(message)

        # This will run in the background, exiting after one minute
        with pubsub_service.subscribe_sse("my_topic") as messages:
            def run():
                for message in messages:
                    handle(message)
            threading.Thread(target=run).start()

            sleep(60)
        """
        open_connections = []
        running = True

        def msgs():
            delay = 1
            while running:
                try:
                    response = requests.get(
                        f"{self.base_url}/subscribe/sse/{self._fully_qualified_topic(topic)}",
                        stream=True,
                    )
                    open_connections.append(response)
                    delay = 1
                    for line in response.iter_lines():
                        if line and line.startswith(b"data:"):
                            payload = line[5:].decode("utf-8").strip()
                            if payload:
                                yield json.loads(payload)
                except requests.exceptions.ConnectionError:
                    logging.warning(
                        "Unable to establish server connection at %s. Sleeping for %ss",
                        self.base_url,
                        delay,
                    )
                    try:
                        open_connections.pop().close()
                    except Exception:
                        pass
                    # Possible misconfiguration. Back off exponentially
                    sleep(delay)
                    delay = min(delay * 2, 60)
                except requests.exceptions.RequestException:
                    logging.warning("Server at %s closed SSE connection", self.base_url)
                    try:
                        open_connections.pop().close()
                    except Exception:
                        pass
                    # Service may have redeployed. Try to reestablish connection quickly
                    sleep(1)

        try:
            yield msgs()
        finally:
            running = False
            for conn in open_connections:
                try:
                    conn.close()
                except Exception:
                    pass

    def publish(self, topic: str, payload: Dict[str, Any]):
        """
        Publish a message to a topic
        """
        ns_topic = self._fully_qualified_topic(topic)
        event = PublishedEvent(topic=ns_topic, payload=payload)
        requests.post(f"{self.base_url}/publish/{ns_topic}", json=event.model_dump())

    def _fully_qualified_topic(self, topic: str) -> str:
        return f"{self.namespace}:{topic}" if self.namespace else topic

    @staticmethod
    def from_env() -> "PubsubService":
        return PubsubService(
            base_url=os.getenv("PUBSUB_URL", "https://nora-pubsub.apps.allenai.org"),
            namespace=os.getenv("PUBSUB_NAMESPACE", os.getenv("ENV", "prod")),
        )


class PublishedEvent(BaseModel):
    topic: str
    payload: Dict[str, Any]



import setuptools

runtime_requirements = ["pydantic>=2,<3", "requests", "boto3", "aws_requests_auth"]

# For running tests, linting, etc
dev_requirements = ["mypy", "pytest", "black", "types-requests"]

setuptools.setup(
    name="nora_lib",
    version="0.4.6rc1",
    description="For making and coordinating agents and tools",
    url="https://github.com/allenai/nora_lib",
    packages=setuptools.find_packages(exclude=(["tests"])),
    install_requires=runtime_requirements,
    package_data={
        "nora_lib": ["py.typed"],
    },
    extras_require={
        "dev": dev_requirements,
    },
    python_requires=">=3.9",
)



from nora_lib.interactions.interactions_service import InteractionsService
from nora_lib.interactions.models import *
import uuid

iservice = InteractionsService("http://localhost:9080")
service_cost = ServiceCost(
    dollar_cost=1,
    service_provider="openai",
    description="Step 1 of PaperQA pipeline",
    tool_name="PaperQA",
    details=[
        LLMCost(token_count=1000, model_name="gpt-turbo"),
        LLMTokenBreakdown(prompt_tokens=100, completion_tokens=100),
    ],
)
step_cost = StepCost(
    actor_id=uuid.uuid4(), message_id="string", service_cost=service_cost
)
iservice.report_cost(step_cost)


class OtherLLMCostDetail(CostDetail):
    coolness_factor: float


service_cost2 = ServiceCost(
    dollar_cost=2,
    service_provider="anthropic",
    description="Step 2 of PaperQA pipeline",
    tool_name="PaperQA",
    details=[OtherLLMCostDetail(coolness_factor=0.5)],
)
step_cost2 = StepCost(
    actor_id=uuid.uuid4(), message_id="string", service_cost=service_cost2
)
iservice.report_cost(step_cost2)
